#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdint.h>
#include <BBB.h>
#include <BBB_GPIO.h>
#include "aki_lcd.h"

static const uint32_t GPIO_REGS[] = {
	0x44E07000,     /* GPIO0 */
	0x4804C000,     /* GPIO1 */
	0x481AC000,     /* GPIO2 */
	0x481AE000      /* GPIO3 */
};

#define GPIO_DATAOUT		0x13C
#define GPIO_DATAIN			0x138
#define GPIO_CLEARDATAOUT	0x190
#define GPIO_SETDATAOUT		0x194

#define LCD_RS_PIN	"P9_12"
#define LCD_E_PIN	"P9_23"
#define LCD_CS1_PIN	"P9_15"
#define LCD_CS2_PIN	"P9_16"

#define LCD_DB7_b	31
#define LCD_DB6_b	30
#define LCD_DB5_b	26
#define LCD_DB4_b	15

#define LCD_DB3_b	5
#define LCD_DB2_b	4
#define LCD_DB1_b	3
#define LCD_DB0_b	2

#define LCD_DB_CLEAR_MASK	((1<<LCD_DB7_b)|(1<<LCD_DB6_b)|(1<<LCD_DB5_b)|(1<<LCD_DB4_b)|\
						     (1<<LCD_DB3_b)|(1<<LCD_DB2_b)|(1<<LCD_DB1_b)|(1<<LCD_DB0_b))

static const char *LCD_DB_PINS[8] 
 = {"P9_22", "P9_21", "P9_18", "P9_17", "P9_24", "P8_14", "P9_11","P9_13"};

static volatile uint32_t	*gpio_reg[4];
static int gpio_fd[4];

static BBB_gpio	*lcd_e;
static BBB_gpio	*lcd_rs;
static BBB_gpio *lcd_cs1;
static BBB_gpio *lcd_cs2;
static BBB_gpio *lcd_db[8];

static int lcd_set_single( BBB_gpio *gpio, uint32_t v )
{
	uint32_t value;
	
	value = 1 << gpio->pin->bit;
	if( v ) 
		gpio_reg[gpio->pin->channel][GPIO_SETDATAOUT/sizeof(uint32_t)] = value;
	else
		gpio_reg[gpio->pin->channel][GPIO_CLEARDATAOUT/sizeof(uint32_t)] = value;
	
	return 0;
}

static int lcd_write_databus( uint32_t data )
{
	uint32_t port_val;
	uint32_t write_val;
	
	write_val = (data & 0x0F) << LCD_DB0_b;	/* 下位4ビット */
	write_val |= ((data & 0x10) >> 4) << LCD_DB4_b;
	write_val |= ((data & 0x20) >> 5) << LCD_DB5_b;
	write_val |= ((data & 0x40) >> 6) << LCD_DB6_b;
	write_val |= ((data & 0x80) >> 7) << LCD_DB7_b;
	
	port_val = gpio_reg[0][GPIO_DATAOUT/sizeof(uint32_t)];
	gpio_reg[0][GPIO_DATAOUT/sizeof(uint32_t)] = port_val | write_val;
}

static int lcd_clear_databus(void)
{
	gpio_reg[0][GPIO_CLEARDATAOUT/sizeof(uint32_t)] = LCD_DB_CLEAR_MASK;
	return 0;
}

static int lcd_write( uint32_t data , uint32_t cs )
{
	/* clear E */
	lcd_set_single(lcd_e, 0 );
	// output cs
	lcd_set_single(lcd_cs1, cs & 0x01 );
	lcd_set_single(lcd_cs2, cs & 0x02 );
	lcd_write_databus( data );
	usleep(10);
	
	// E
	lcd_set_single(lcd_e, 1);
	usleep(10);
	lcd_set_single(lcd_e, 0);
	usleep(10);
	
	// clear DATA BUS
	lcd_clear_databus();
	lcd_set_single(lcd_cs1, 0);
	lcd_set_single(lcd_cs2, 0);
	return 0;
}

int lcd_write_cmd( uint8_t cmd, uint8_t cs )
{
	lcd_set_single(lcd_rs, 0 );
	usleep(10);
	lcd_write(cmd, cs);
	lcd_set_single(lcd_rs, 0 );
	usleep(10);
	return 0;
}

int lcd_write_display_data( uint8_t data, uint8_t page, uint8_t cs )
{
	lcd_write_cmd( 0xB8 | page, cs );	/* set page */	
	lcd_set_single(lcd_rs, 1);
	lcd_write( data, cs );
    lcd_set_single(lcd_rs, 0);
	usleep(10);
	return 0;
}

static int lcd_init_databus( void )
{
	int i;
	
	for( i = 0; i < 4; i++ ) {
		gpio_fd[i] = open("/dev/mem", O_RDWR | O_SYNC );
		if( gpio_fd[i] <= 0 ) {
			return -1;
		}
		gpio_reg[i] = (uint32_t *)mmap( NULL,
							  0xFFF,
							  PROT_READ|PROT_WRITE,
							  MAP_SHARED,
							  gpio_fd[i],
							  GPIO_REGS[i]
						);
		if( gpio_reg <= 0 ){
			close(gpio_fd[i]);
			return -1;
		}
	}
	lcd_clear_databus();
	
	return 0;
}

int lcd_init( void )
{
	int i;
	int f_error = 0;
	
	lcd_e = BBB_open_gpio( LCD_E_PIN );
	lcd_rs = BBB_open_gpio( LCD_RS_PIN );
	lcd_cs1 = BBB_open_gpio( LCD_CS1_PIN );
	lcd_cs2 = BBB_open_gpio( LCD_CS2_PIN );
	
	for( i=0; i<8 ;i++ ) {
		lcd_db[i] = NULL;
		lcd_db[i] = BBB_open_gpio((char *)LCD_DB_PINS[i]);
		if( lcd_db[i] == NULL ) {
			printf("%s error\n", LCD_DB_PINS[i]);
			f_error = 1;
		}
	}
	if( lcd_init_databus() < 0 ) f_error = 1;
	
	if( f_error || (lcd_e == NULL) || (lcd_rs == NULL) || (lcd_cs1 == NULL) || (lcd_cs2 == NULL) ) {
		if( lcd_e )   BBB_close_gpio(lcd_e);
		if( lcd_rs )  BBB_close_gpio(lcd_rs);
		if( lcd_cs1 ) BBB_close_gpio(lcd_cs1);
		if( lcd_cs2 ) BBB_close_gpio(lcd_cs2);
		for( i=0; i<8; i++ ) {
			if( lcd_db[i] ) BBB_close_gpio(lcd_db[i] );
		}
		return -1;
	}
	lcd_e->set_direction(lcd_e, OUT);
	lcd_rs->set_direction(lcd_rs, OUT);
	lcd_cs1->set_direction(lcd_cs1, OUT);
	lcd_cs2->set_direction(lcd_cs2, OUT);
	
	for( i=0; i<8 ;i++ ) {
		lcd_db[i]->set_direction(lcd_db[i], OUT );
	}
	
	// clear Bus
	lcd_set_single(lcd_e, 0);
	lcd_set_single(lcd_rs, 0);
	lcd_set_single(lcd_cs1,0);
	lcd_set_single(lcd_cs2,0);
	lcd_clear_databus();

	usleep( 30 * 3000 );		/* wait 30mS */
	 
	lcd_write_cmd( 0xC0, 3 );	/* Set Display Start Line */
	lcd_write_cmd( 0x3F, 3 );	/* Display on */

    return 0;
}


int lcd_close(void)
{
	int i;
	
	for( i =0; i < 4; i++ ) {
		munmap((void *) gpio_reg[i], 0xFFF );
		close(gpio_fd[i]);
	}
	
	BBB_close_gpio( lcd_e );
	BBB_close_gpio( lcd_rs );
	BBB_close_gpio( lcd_cs1 );
	BBB_close_gpio( lcd_cs2 );
	
	for( i=0; i<8 ;i++ ) {
		BBB_close_gpio( lcd_db[i] );
	}
	
	return 0;
}

